@@ -7,6 +7,7 @@
 import calendar
 import datetime
 import itertools
+import re
 import sys
 import types
 import warnings
@@ -31,7 +32,14 @@ MONTHS = dict(
 
 MULTI_KEYWORDS = set(["reference"])
 MULTI_QUALIFIERS = set("""
-    allele db_xref ec_number go_component go_process go_function transl_except
+    allele
+    db_xref
+    ec_number
+    function
+    go_component go_process go_function
+    mobile_element
+    old_locus_tag
+    transl_except
 """.split())
 #
 # Parse Actions
@@ -128,10 +136,10 @@ pid = Literal("PID").suppress() + Word(alphanums)
 dbsource = Literal("DBSOURCE").suppress() + dedented
 
 # DBLINK
-dbln_type = Or(map(Literal, ["Project", "Trace Assembly Archive"]))("type")
+dbln_type = Or(map(Literal, ["Project", "Trace Assembly Archive"]))
 dbln_type = dbln_type.setParseAction(lambda t: t[0].lower())
-dbln_list = Group(delimitedList(Word(alphanums + "_")))("identifiers")
-dblink = Group(dbln_type + sup_colon + dbln_list)("dblink")
+dbln_list = Group(delimitedList(Word(alphanums + "_")))
+dblink = Literal("DBLINK") + Group(dbln_type + sup_colon + dbln_list)
 
 # PROJECT
 project = Literal("PROJECT").suppress() + normalized
@@ -163,12 +171,15 @@ notes = Word(''.join(p for p in printables if p not in "()"))("notes")
 refloc = sup_lparen + Optional(position) + Optional(notes) + sup_rparen
 refdesc = integer("refid") + Group(refloc)
 authors = Group(Literal("AUTHORS") + keywordnorm)
+consrtm = Group(Literal("CONSRTM") + keywordnorm)
 title = Group(Literal("TITLE") + keywordnorm)
 journal = Group(Literal("JOURNAL") + keywordnorm)
 medline = Group(Literal("MEDLINE") + Word(nums, exact=8)("medline"))
 pubmed = Group(Literal("PUBMED") + Word(nums)("pubmed"))
 remark = Group(Literal("REMARK") + keywordnorm)
-refsubs = OneOrMore(authors ^ title ^ journal ^ medline ^ pubmed ^ remark)
+refsubs = OneOrMore(
+    authors ^ consrtm ^ title ^ journal ^ medline ^ pubmed ^ remark
+)
 reference = Literal("REFERENCE").suppress() + refdesc + refsubs
 
 # PRIMARY
@@ -198,31 +209,34 @@ join = Group(Literal("join") + args)("join")
 order = Group(Literal("order") + args)("order")
 referenceloc = Group(veracc + sup_colon + location)("reference")
 location << (
-    complement | join | order | bond | gap | referenceloc \
-        | span | site | choice | index | one_of
+    span | complement | join | order | bond | gap | referenceloc \
+        | site | choice | index | one_of
 )
 
 # FEATURES
-featcont = White(exact=21).suppress() + restOfLine + LineEnd()
-feature = Combine(
-    White(exact=5).suppress() + restOfLine + LineEnd() + ZeroOrMore(featcont),
-    adjacent=False
-)
-featuresection = OneOrMore(feature)
+#
+# Hand coded feature spliting to speed up parsing.
+#
+# featcont = White(exact=21).suppress() + restOfLine + LineEnd()
+# feature = Combine(
+#    White(exact=5).suppress() + restOfLine + LineEnd() + ZeroOrMore(featcont),
+#     adjacent=False
+# )
+# featuresection = OneOrMore(feature)
 
-featname = Word(alphanums + "_'")
-featlocword = Word(''.join(filter(lambda p: p != "/", printables)))
-featloc = Combine(OneOrMore(featlocword), adjacent=False)
-qualname = Word(''.join(filter(lambda p: p != "=", printables)))
-qualquoted = Combine(
-    Suppress('"') + Regex(r'[^"]*') + Suppress('"'),
-    adjacent=True
-)
-qualunquoted = Word(filter(lambda p: p != '"', printables))
-qualvalue = Suppress("=") + (qualquoted | qualunquoted) + LineEnd().suppress()
-qualnovalue = LineEnd().suppress()
-qualifier = Group(Suppress("/") + qualname + (qualvalue | qualnovalue))
-featgroup = featname + featloc + ZeroOrMore(qualifier)
+# featname = Word(alphanums + "_'")
+# featlocword = Word(''.join(filter(lambda p: p != "/", printables)))
+# featloc = Combine(OneOrMore(featlocword), adjacent=False)
+# qualname = Word(''.join(filter(lambda p: p != "=", printables)))
+# qualquoted = Combine(
+#     Suppress('"') + Regex(r'[^"]*') + Suppress('"'),
+#     adjacent=True
+# )
+# qualunquoted = Word(filter(lambda p: p != '"', printables))
+# qualvalue = Suppress("=") + (qualquoted | qualunquoted) + LineEnd().suppress()
+# qualnovalue = LineEnd().suppress()
+# qualifier = Group(Suppress("/") + qualname + (qualvalue | qualnovalue))
+# featgroup = featname + featloc + ZeroOrMore(qualifier)
 
 # CONTIG
 contig = Literal("CONTIG") + location
@@ -265,6 +279,16 @@ def _k_pid(data):
 def _k_dbsource(data):
     return dbsource.parseString(data)[0].strip()
 
+def _k_dblink(data):
+    ret = []
+    res = dblink.parseString(data)
+    for group in res:
+        ret.append({
+            "type": group[0],
+            "accessions": group[1]
+        })
+    return ret
+
 def _k_segment(data):
     res = segment.parseString(data)
     return {"position": res.position, "of": res.total}
@@ -404,67 +428,45 @@ def _l_referenceloc(t):
         "segment": t[0][1]
     }
 
-loctypenames = """
-    index bond one_of span site choice gap complement join order referenceloc
-""".split()
+def _mk_locs():
+    lnames = """
+        index bond one_of span site choice gap
+        complement join order referenceloc
+    """.split()
+    for ln in lnames:
+        elem = globals()[ln].setParseAction(globals()["_l_%s" % ln])
+        globals()[ln] = elem
+_mk_locs()
 
-for l in loctypenames:
-    locals()[l].setParseAction(locals()["_l_%s" % l])
 
 
 #
 # Qualifier Parsers
 #
+_qual_default = lambda d: ' '.join(d.strip('"').split())
 
 def _q_transl_except(d):
     res = modification.parseString(d)
     return {"location": res[0], "acid": res[1]}
 
-_qual_default = lambda d: d.strip('"')
-_q_allele = _qual_default
 _q_anticodon = _q_transl_except
-_q_bond_type = _qual_default
-_q_chromosome = _qual_default
-_q_clone = _qual_default
-_q_clone_lib = _qual_default
-_q_coded_by = _qual_default
 _q_codon_start = lambda d: int(d)
-_q_country = _qual_default
-_q_cultivar = _qual_default
-_q_db_xref = _qual_default
-_q_dev_stage = _qual_default
-_q_ec_number = _qual_default
-_q_evidence = _qual_default
-_q_exception = _qual_default
-_q_function = lambda d: " ".join(d.split())
-_q_gene = _qual_default
-_q_go_component = _qual_default
-_q_go_function = _qual_default
-_q_go_process = _qual_default
-_q_isolate = _qual_default
-_q_map = _qual_default
-_q_mol_type = _qual_default
-_q_note = lambda d: " ".join(d.split())
-_q_ncrna_class = _qual_default
 _q_number = lambda d: int(d)
-_q_organelle = _qual_default
-_q_organism = _qual_default
-_q_product = _qual_default
-_q_protein_id = _qual_default
 _q_pseudo = lambda d: True
-_q_region_name = _qual_default
 _q_ribosomal_slippage = lambda d: True
-_q_rpt_family = _qual_default
 _q_selenocysteine = lambda d: True
-_q_site_type = _qual_default
-_q_standard_name = _qual_default
-_q_strain = _qual_default
-_q_tissue_type = _qual_default
-_q_transcript_id = _qual_default
-_q_transl_table = _qual_default
-_q_translation = lambda d: ''.join(d.split())
+_q_translation = lambda d: ''.join(d.strip('"').split())
 _q_virion = lambda d: True
 
+known_qualifiers = set("""
+allele bond_type chromosome citation clone clone_lib coded_by codon_recognized
+country cultivar db_xref dev_stage direction ec_number evidence exception
+experiment function gene gene_synonym go_component go_function go_process
+inference isolate locus_tag map mobile_element mol_type note ncrna_class
+old_locus_tag operon organelle organism product protein_id region_name
+rpt_family site_type standard_name strain sub_strain tissue_type transcript_id
+transl_table
+""".split())
 
 def _add_qualifier(s, quals, name, data):
     if name not in MULTI_QUALIFIERS:
@@ -515,29 +517,99 @@ def parse_keywords(data):
 def parse_location(data):
     return location.parseString(data)[0]
 
+def parse_qual_name(data, pos):
+    while pos < len(data) and data[pos] == " ":
+        pos += 1
+    if data[pos] != "/":
+        raise ParseException(data, len(data), "Incorrect qualifier name start.")
+    for i in range(pos+1, len(data)):
+        if data[i] == "=" or data[i] == " ":
+            return (data[pos+1:i], i)
+    return (data[pos+1:], len(data))
+
+def parse_qual_value(data, pos):
+    if pos >= len(data):
+        return (None, pos)
+    if data[pos] == " ":
+        return (None, pos+1)
+    if data[pos] != "=":
+        raise ParseException(data, len(data), "Incorrect qualifiver val start.")
+    if data[pos+1] == '"':
+        off, delim = 1, '"'
+    else:
+        off, delim = 0, ' '
+    for i in range(pos+1+off, len(data)):
+        if data[i] == delim:
+            return (data[pos+1:i+off], i+off)
+    return (data[pos+1:], len(data))
+
+def parse_qualifiers(data):
+    if not data:
+        return
+    pos = 0
+    while pos < len(data):
+        (qname, pos) = parse_qual_name(data, pos)
+        (qval, pos) = parse_qual_value(data, pos)
+        yield (qname, qval)
+
+FEAT_GROUP = re.compile("^(\S+)\s+([^/]+)(?:\s+(.*))?$")
 def parse_feature(data):
-    res = featgroup.parseString(data)
+    data = ' '.join(data.split())
+    match = FEAT_GROUP.match(data)
+    if not match:
+        raise ValueError(data)
+    feattype = match.group(1)
+    featloc = location.parseString(match.group(2))
     props = {}
-    for q in res[2:]:
-        qname = q[0].lower()
-        value = q[1] if len(q) > 1 else None
-        func = globals().get("_q_%s" % qname)
+    glbls = globals()
+    for (qname, qval) in parse_qualifiers(match.group(3)):
+        qname, qval = qname.lower(), qval.strip('"') if qval else None
+        func = glbls.get("_q_%s" % qname)
         if func is None:
-            warnings.warn("No parser for qualifier: %s" % qname)
-            _add_qualifier(data, props, qname, value)
+            if qname not in known_qualifiers:
+                warnings.warn("No parser for qualifier: %s" % qname)
+            _add_qualifier(data, props, qname, qval)
         else:
-            _add_qualifier(data, props, qname, func(value))
+            _add_qualifier(data, props, qname, func(qval))
     return {
-        "type": res[0],
-        "location": parse_location(res[1]),
+        "type": match.group(1),
+        "location": parse_location(match.group(2)),
         "properties": props
     }
+    #res = featgroup.parseString(data)
+    #props = {}
+    #for q in res[2:]:
+    #    qname = q[0].lower()
+    #    value = q[1] if len(q) > 1 else None
+    #    func = globals().get("_q_%s" % qname)
+    #    if func is None:
+    #        warnings.warn("No parser for qualifier: %s" % qname)
+    #        _add_qualifier(data, props, qname, value)
+    #    else:
+    #        _add_qualifier(data, props, qname, func(value))
+    #return {
+    #    "type": res[0],
+    #    "location": parse_location(res[1]),
+    #    "properties": props
+    #}
 
+NEXT_FEAT = re.compile("^\s{5}\S")
 def parse_features(data):
     ret = []
-    for feat in featuresection.parseString(data):
-        ret.append(parse_feature(feat))
+    curr = []
+    for line in Reader.from_string(data):
+        if NEXT_FEAT.match(line):
+            if len(curr) > 0:
+                ret.append(parse_feature(''.join(curr)))
+            curr = [line]
+        else:
+            curr.append(line)
+    if len(curr):
+        ret.append(parse_feature(''.join(curr)))
     return ret
+    #for feat in featuresection.parseString(data):
+    #    ret.append(parse_feature(feat))
+    #return ret
 
 def parse_contig(data):
     return contig.parseString(data)[0]
@@ -563,8 +635,8 @@ class Reader(object):
             prev = 0
             next = data.find("\n")
             while next >= 0:
-                yield data[prev:next]
-                prev, next = next, data.find("\n", next+1)
+                yield data[prev:next+1]
+                prev, next = next+1, data.find("\n", next+1)
             yield data[prev:]
         return Reader(_iter(strdata))
 
